/*
	File:		USBTabletModule.c

	Contains:	HID Module for USB Tablet

	Version:	xxx put version here xxx

	Copyright:	 1997-1998 by Apple Computer, Inc., all rights reserved.

*/

#include <Types.h>
#include <Devices.h>
#include <processes.h>
#include <DriverServices.h>
#include <USB.h>

#include "USBTabletModule.h"

usbTabletPBStruct myTabletPB;

void InitParamBlock(USBDeviceRef theDeviceRef, USBPB * paramblock)
{
	paramblock->usbReference = theDeviceRef;
	paramblock->pbVersion = kUSBCurrentPBVersion;
	paramblock->usbStatus = noErr;
	paramblock->usbBuffer = nil;		
	paramblock->usbActCount = 0;
	paramblock->usbReqCount = 0;
	paramblock->usbFlags = 0;
	paramblock->usbOther = 0;
}



Boolean immediateError(OSStatus err)
{
	return((err != kUSBPending) && (err != noErr) );
}

void InitiateTransactionProc(USBPB *pb)
{
register usbTabletPBStruct *pTabletPB;
OSStatus myErr;

	pTabletPB = (usbTabletPBStruct *)(pb);
	pTabletPB->transDepth++;
	if (pTabletPB->transDepth < 0)
	{
		USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: transDepth < 0 (initiation)", pTabletPB->pb.usbRefcon );
	}
	
	if (pTabletPB->transDepth > 1)
	{
		USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: transDepth > 1 (initiation)", pTabletPB->pb.usbRefcon );
	}
	
	switch(pTabletPB->pb.usbRefcon & ~kRetryTransaction)
	{
		case kFindInterface:
			InitParamBlock(pTabletPB->deviceRef, &pTabletPB->pb);
			pTabletPB->pb.usbWIndex = 0;
			pTabletPB->pb.usbWValue = 0;
			pTabletPB->pb.usbClassType = kUSBHIDInterfaceClass;
			pTabletPB->pb.usbSubclass = kUSBBootInterfaceSubClass;
			pTabletPB->pb.usbProtocol = kUSBMouseInterfaceProtocol;
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBFindNextInterface(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->pb.usbReference, kUSBInternalErr, "\pTablet Driver: USBFindNextInterface (immediate error)", myErr);
				pTabletPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kOpenDevice:
			InitParamBlock(pTabletPB->deviceRef, &pTabletPB->pb);
			pTabletPB->pb.usbWValue = pTabletPB->configurationNumber;
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBOpenDevice(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->pb.usbReference, kUSBInternalErr, "\pTablet Driver: USBOpenDevice (immediate error)", myErr);
				pTabletPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kNewInterfaceRef:
			InitParamBlock(pTabletPB->deviceRef, &pTabletPB->pb);
			pTabletPB->pb.usbWIndex = pTabletPB->interfaceNumber;
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBNewInterfaceRef(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->pb.usbReference, kUSBInternalErr, "\pTablet Driver: USBNewInterfaceRef (immediate error)", myErr);
				pTabletPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kConfigureInterface:
			InitParamBlock(pTabletPB->interfaceRef, &pTabletPB->pb);
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBConfigureInterface(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->pb.usbReference, kUSBInternalErr, "\pTablet Driver: USBConfigureInterface (immediate error)", myErr);
				pTabletPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kSetProtocol:
			InitParamBlock(pTabletPB->interfaceRef, &pTabletPB->pb);
			
			pTabletPB->pb.usbBMRequestType = USBMakeBMRequestType(kUSBOut, kUSBClass, kUSBInterface);			
			pTabletPB->pb.usbBRequest = kHIDRqSetProtocol;
			pTabletPB->pb.usbWValue = kHIDReportProtocolValue; 
			pTabletPB->pb.usbWIndex = pTabletPB->interfaceNumber;
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBDeviceRequest(&pTabletPB->pb);
			if (immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: kHIDRqSetProtocol (immediate error)", myErr);
				pTabletPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kGetReportDescriptor:
			InitParamBlock(pTabletPB->deviceRef, &pTabletPB->pb);
			pTabletPB->pb.usbBMRequestType = USBMakeBMRequestType(kUSBIn, kUSBStandard, kUSBInterface);
			pTabletPB->pb.usbBRequest = kUSBRqGetDescriptor;
			pTabletPB->pb.usbWValue = (0x22 << 8) + 1; 			// get the first report descriptor
			pTabletPB->pb.usbWIndex = pTabletPB->interfaceNumber;
			pTabletPB->pb.usbReqCount = 0x66;
			pTabletPB->pb.usbBuffer = pTabletPB->hidReportDescriptor;
			
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBDeviceRequest(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->pb.usbReference, kUSBInternalErr, "\pTablet Driver: kGetReportDescriptor - immediate error", myErr);
			}
			break;
		
		case kFindInterruptPipe:
			InitParamBlock(pTabletPB->interfaceRef, &pTabletPB->pb);
			pTabletPB->pb.usbFlags = kUSBIn;
			pTabletPB->pb.usbClassType = kUSBInterrupt; 
			
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBFindNextPipe(&pTabletPB->pb);
			if (immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: kFindInterruptPipe (immediate error)", myErr);
				pTabletPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kReadInterruptPipe:
			InitParamBlock(pTabletPB->pipeRef, &pTabletPB->pb);

			pTabletPB->pb.usbBuffer = (Ptr)pTabletPB->hidReport;
			pTabletPB->pb.usbReqCount = 7;
			pTabletPB->pb.usbWIndex = pTabletPB->interfaceNumber;	
			
			pTabletPB->pb.usbRefcon |= kCompletionPending;
			pTabletPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBIntRead(&pTabletPB->pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: kReadInterruptPipe (ImmediateError)", myErr);
			}
			break;
			
		default:
			USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: Transaction initiated with bad refcon value", pTabletPB->pb.usbRefcon);
			pTabletPB->pb.usbRefcon = kUndefined + kReturnFromDriver;
			break;
	}
	
// At this point the control is returned to the system.  If a USB transaction
// has been initiated, then it will call the Complete procs
// (below) to handle the results of the transaction.
}


void TransactionCompletionProc(USBPB *pb)
{
register usbTabletPBStruct *pTabletPB;
unsigned char	* errstring;
USBPipeState 	pipeState;

	pTabletPB = (usbTabletPBStruct *)(pb);
	pTabletPB->transDepth--;
	if (pTabletPB->transDepth < 0)
	{
		USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: transDepth < 0 (completion)", pTabletPB->pb.usbRefcon );
	}
	
	if (pTabletPB->transDepth > 1)
	{
		USBExpertFatalError(pTabletPB->deviceRef, kUSBInternalErr, "\pTablet Driver: transDepth > 1 (completion)", pTabletPB->pb.usbRefcon );
	}
	
	if(pTabletPB->pb.usbStatus != noErr)														// was there an error?
	{
		switch(pTabletPB->pb.usbRefcon & 0x0fff)												// yes, so show where the error occurred
		{
			case kFindInterface:  				errstring = "\pTablet Driver: Error during kFindInterface"; break;
			case kOpenDevice:					errstring = "\pTablet Driver: Error during kOpenDevice"; break;
			case kNewInterfaceRef:				errstring = "\pTablet Driver: Error during kNewInterfaceRef"; break;
			case kConfigureInterface:  			errstring = "\pTablet Driver: Error during kConfigureInterface"; break;
			case kSetProtocol: 	 				errstring = "\pTablet Driver: Error during kSetProtocol"; break;
			case kGetReportDescriptor: 	 		errstring = "\pTablet Driver: Error during kGetReportDescriptor"; break;
			case kFindInterruptPipe: 	 		errstring = "\pTablet Driver: Error during kFindInterruptPipe"; break;
			case kReadInterruptPipe: 	 		errstring = "\pTablet Driver: Error during kReadInterruptPipe"; break;
			default: 	 						errstring = "\pTablet Driver: Error occurred, but state is unknown"; break;
		};
		USBExpertFatalError(pTabletPB->deviceRef, pTabletPB->pb.usbStatus, errstring, (pTabletPB->pb.usbRefcon & 0x0fff));
		
		pTabletPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);				// set up to retry the transaction
		pTabletPB->pb.usbRefcon |= kRetryTransaction;
		pTabletPB->retryCount--;
		
		if ((!pTabletPB->retryCount)	|| (pTabletPB->pb.usbStatus == kUSBAbortedError))	// have we exhausted the retries?
		{																				// or received an abort?
			USBExpertStatus(pTabletPB->deviceRef, "\pTablet Driver: Pipe abort or unable to recover from error", pTabletPB->deviceRef);
			pTabletPB->pb.usbRefcon = kReturnFromDriver;									// if so, just exit.
		}
		else																			// if it didn't abort and there's retries left, then...
		{
			if (pTabletPB->pipeRef)														// check if the pipe is open.
			{
				USBGetPipeStatusByReference(pTabletPB->pipeRef, &pipeState);				// yes, so what it's state?
				if (pipeState != kUSBActive)											// if it's not active, try to clear it.  It might be stalled...
				{
					USBExpertStatus(pTabletPB->deviceRef, "\pTablet Driver: Pipe is open and stalled, clearing stall...", pTabletPB->deviceRef);
					USBClearPipeStallByReference(pTabletPB->pipeRef);
				}
			}
		}
	}
	else
	{
		pTabletPB->pb.usbRefcon &= ~kRetryTransaction;
		pTabletPB->retryCount = kTabletRetryCount;
	}

	if (pTabletPB->pb.usbRefcon & kCompletionPending)			 
	{												
		pTabletPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);
		switch(pTabletPB->pb.usbRefcon)
		{
			case kFindInterface:
				if ((pTabletPB->pb.usbClassType == kUSBHIDInterfaceClass) && 
					(pTabletPB->pb.usbSubclass == kUSBBootInterfaceSubClass) &&
					(pTabletPB->pb.usbProtocol == kUSBMouseInterfaceProtocol))
				{
					pTabletPB->pb.usbRefcon = kOpenDevice;
					pTabletPB->interfaceNumber = pTabletPB->pb.usbWIndex;
					pTabletPB->configurationNumber = pTabletPB->pb.usbWValue;
				}
				else
				{
					pTabletPB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kOpenDevice:
				pTabletPB->pb.usbRefcon = kNewInterfaceRef;
				break;
				
			case kNewInterfaceRef:
				pTabletPB->interfaceRef = pTabletPB->pb.usbReference;
				pTabletPB->pb.usbRefcon = kConfigureInterface;
				break;
			
			case kConfigureInterface:
				pTabletPB->pb.usbRefcon = kGetReportDescriptor;
				break;
			
			case kGetReportDescriptor:
				pTabletPB->pb.usbRefcon = kFindInterruptPipe;
				break;
				
			case kFindInterruptPipe:
				if ((pTabletPB->pb.usbClassType == kUSBInterrupt) && 
					(pTabletPB->pb.usbFlags == kUSBIn))
				{
					pTabletPB->hidReportSize = pTabletPB->pb.usbWValue;
					pTabletPB->pipeRef = pTabletPB->pb.usbReference;
					pTabletPB->pb.usbRefcon = kReadInterruptPipe;
				}
				else
				{
					pTabletPB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kReadInterruptPipe:
				//DebugStr("\pTablet read completed");
				ProcessInterruptReport(pTabletPB->refcon, pTabletPB->hidReport);
				pTabletPB->pb.usbRefcon = kReadInterruptPipe;
				break;

		}
	}
	if (!(pTabletPB->pb.usbRefcon & kReturnFromDriver))
		InitiateTransactionProc(pb);
}

